/*
   File: cdl3rts.c
   Defines CDL3 runtime system routines

   Copyright (C) 2000 C.H.A. Koster
  
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA

   CVS ID: "$Id$"
*/

/* Check to include config.h */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <signal.h>

/* Unix only includes to use times () for resource usage */
#ifndef WIN32
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>	/* for sysconf() */
#endif

/* Local includes */
#include "cdl3rts.h"

#define MAXSIZE 50
#define CHUNK 20

extern void (*errhandler)();

long modulenr = -1;
long linenr = 0;
int received_signal = 0;

#ifndef CDLASM
#define STATIC static
#else
#define STATIC /**/
#endif

STATIC value freelists[MAXSIZE];
STATIC value backuplist[MAXSIZE];

long attaches = 0;
long detaches = 0;
long frees = 0;
STATIC long used = 0;
STATIC long reused = 0;
STATIC long requested = 0;
STATIC long freed = 0;

long saved_calls = 0;
long array_ref = 0;
long array_done = 0;
long chunksize = 1024 * 1024;	/* Size of data chunks */

STATIC value remerge = NULL;
STATIC value maxlist = NULL;
STATIC value backupmax = NULL;
STATIC long totsize = 0;
STATIC value recycle = NULL;
STATIC value allmem = NULL;

/* Status information */
long CDL3Usage = 0;
long CDL3FancyErrors = 0;
long CDL3Compiler = 0;

/* Export values of original argument list */
int argument_count;
char **arguments;

#ifndef CDLASM
static value mmalloc (int size)
{ value v = (value) calloc (1, size * sizeof (value));
  if (v == NULL)
    { fprintf (stderr, "Out of memory... (Allocating %d).\n", size);
      quit (1);
    };
  return (v);
}
#endif

#ifndef CDLASM
value getmem (long size)
{ value v;
  if (size >= MAXSIZE)
    { /* large structure malloc */
      v = mmalloc (size);
      used += (long) (size * sizeof (value));
    }
  else if (freelists[size] != NULL)
    { /* reuse */
      v = freelists[size];
      freelists[size] = (value) (*v);
      reused += (long) (size * sizeof (value));
    }
  else
    { /* MULTIPLE MALLOC */
      int i;

      v = (value) mmalloc (size * CHUNK);
      used += size * sizeof (value) * CHUNK;

      for (i = 0; i < CHUNK; i++, v += size)
        { *v = (long) (freelists[size]);
          freelists[size] = v;
        };

      v = freelists[size];
      freelists[size] = (value) (*v);
    };

  requested += (long) (size * sizeof (value));
  return (v);
}

void freemem (value v, long size)
{ if (size < MAXSIZE)
    { *v = (long) (freelists[size]);
      freelists[size] = v;
    }
  else free (v);

  freed += (long) (size * sizeof (value));
}
#endif /* CDLASM */

value ctext (char *t)
{ int len = strlen (t);
  value v;
#ifndef CDLASM
  v = getmem (3);
  v[0] = Tag (0, T_TEXT);
  v[1] = len;
  v[2] = (long) getmem (Length (len+1));
  strcpy (V_TEXT(v), t);
#else
  v = getmem (2 + Length (len+1));
  v[0] = Tag (0, T_TEXT);
  v[1] = len;
  strcpy (v + 2,t);
#endif
  return (v);
}

value concat (value s1, value s2)
{ int l1 = TextLength (s1);
  int l2 = TextLength (s2);
  value r;
  char *v;

#ifndef CDLASM
  r = getmem (3);
  r[0] = Tag (0, T_TEXT);
  r[1] = l1+l2;
  r[2] = (long)getmem(Length(l1+l2+1));
#else
  r = getmem (2+l1+l2);
  r[0] = Tag (0, T_TEXT);
  r[1] = l1+l2;
#endif
  v = V_TEXT(r);
  strncpy (v, V_TEXT (s1), l1);
  strncpy (v + l1, V_TEXT (s2), l2);
  v[l1+l2] = '\0';
  
  return (r);
}

/*
   Code to give fancy errors on receiving a SIGINT
   Just to brighten up dull days while coding
*/
static char *FancyErrors[38] =
{ "You burned CDL3 to death.",
  "Your lightning bolt kills CDL3.",
  "You chill CDL3, remember to put flowers on its grave.",
  "CDL3 is dissolved by your goodness.",
  "CDL3 crumples as you kill it by draining its energy.",
  "Your fireball hits CDL3 with full force, causing immediate death.",
  "CDL3 reels from the sheer power of your spell, and is destroyed.",
  "Your lightning bolt shatters CDL3 to pieces.",
  "The magic missile tears away the remaining life of CDL3.",
  "CDL3 dies while looking rather shocked.",
  "CDL3 makes a strange sound but is suddenly very silent, as you place your dagger in its back.",
  "Your kick at CDL3's face splits its head open - yummy.",
  "CDL3 doesn't recover from your bash ! - CDL3 is dead.",
  "Your light wounds are enough to send CDL3 over the brink....",
  "You critically kill CDL3.",
  "Your flaming power chars CDL3 to death!",
  "CDL3 is dissolved by your evilness.",
  "CDL3 is dissolved into a sticky ooze.",
  "CDL3 is frozen to death by your cone of cold.",
  "CDL3 is whacked by your swarm of meteors.",
  "CDL3 is ripped apart by your storm of ice.",
  "CDL3 dies instantly from the power of your holy word.",
  "CDL3 dies instantly from the power of your unholy word.",
  "CDL3 dies instantly from the power of your word.",
  "CDL3 is struck by your ray of sunlight, and destroyed instantly.",
  "CDL3 is killed instantly by the dreaded quivering palm.",
  "CDL3 is killed by your fiery breath.",
  "CDL3 is killed by your gas breath.",
  "CDL3 is frozen to death by your frost breath.",
  "CDL3 melts into a pile of stick ooze before your eyes.",
  "CDL3 is killed by your lightning breath.",
  "You punch CDL3's head with a blow, that crushes its skull.",
  "You punch CDL3 in its kidneys resulting in its immediate death.",
  "You bludgeon CDL3 to death.",
  "You succesfully pierce CDL3. The dead body falls to the ground.",
  "You pierce CDL3's heart, you heartbreaker you.",
  "You beautifully slash CDL3 into two parts - both quite dead.",
  "Your slash at CDL3 results in blood spurting all over, and its immediate death."
};

static void FancyError (void)
{ srand ((unsigned int) time (NULL));
  fprintf (stderr, "%s\n"
	   "CDL3 is dead! R.I.P.\n"
	   "Your blood freezes as you hear CDL3's death cry.\n", FancyErrors[rand () % 38]);
}

/* signal handler for control-C & segfaults */
#ifndef NSIG
#define NSIG 32
#endif
static char *signal_names[] =
{ "BAD", "HUP", "INT", "QUIT", "ILL", "TRAP", "IOT",
  "EMT", "FPE", "KILL", "BUS", "SEGV", "SYS", "PIPE",
  "ALRM", "TERM", "URG", "STOP", "TSTP", "CONT", "CHLD",
  "TTIN", "TTOU", "IO", "XCPU", "XFSZ", "VTALRM", "PROF",
  "WINCH", "LOST", "USR1", "USR2"
};

static void sigerr (int sig)
{ fflush (stdout);
  fflush (stderr);

  /* We should be able to determine the filename from the module nr */
  if (modulenr > 0)
    fprintf (stderr, "SIG%s interrupt in line %ld of %s\n",
	     signal_names[sig], linenr, "<unknown>");
    /* Check for SIGINT and fancy errors */
  else if ((sig == 2) && CDL3FancyErrors) FancyError ();
  else fprintf (stderr, "SIG%s interrupt\n", signal_names[sig]);

  received_signal = sig;
  quit (1);
}

void start_rts (int argc, char *argv[])
{ int i;

  /* Save arguments and initialize free lists */
  argument_count = argc;
  arguments = argv;
  for (i = 0; i < MAXSIZE; i++)
    freelists[i] = NULL;
#ifdef NEW_GETMEM
  freelist = NULL;
#endif

  /* Set internal variables from the environment */
  if (getenv ("CDL3USAGE") != NULL)
    { CDL3Usage = 1;
      chunksize = 1024 * 16;
    };

  if (getenv ("CDL_FANCY_ERRORS") != (char *) NULL)
    CDL3FancyErrors = 1;

  /* if our executable contains the string cdl, assume it is the compiler */
  if (strncmp (argv[0], "cdl", 3) == 0)
    CDL3Compiler = 1;

  if (strstr (argv[0], "/cdl") != (char *) NULL)
    CDL3Compiler = 1;

  /* Install signal handlers */
  for (i = 1; i < NSIG; i++)
    switch (i) {
#ifndef WIN32
      case SIGCHLD:
      case SIGTSTP:
      case SIGCONT:
      case SIGWINCH:
      case SIGPROF:
        break;
#endif
    default: (void) signal (i, sigerr);
    };
}

void abort_rts (long line, char* module_name)
{ fflush (stdout);
  fflush (stderr);

  if (CDL3FancyErrors && CDL3Compiler)
    fprintf (stderr, "Internal compiler error in line %ld of module %s\n"
             "Please contact the development team.\n", line, module_name);
  else fprintf (stderr, "Aborted in line %ld of %s\n", line, module_name);

#ifdef ABORT_BY_ASSERT
  assert(0);
#endif

  quit (2);
}

void stop_rts ()
{ quit (0);
}

extern int cdl_main ();

int main (int argc, char *argv[])
{ start_rts (argc, argv);
  cdl_main ();
  stop_rts ();
  return (0);
}

void quit (int e)
{ fflush (stdout);
  fflush (stderr);
  if ((e>0) && (errhandler != NULL))
    errhandler ();

#ifndef WIN32
  if (CDL3Usage)
    { struct tms t;
#ifdef EXTRAGC
      extern char _edata;
      used = ((char *)sbrk(0))-&_edata;
#endif
      int clk_tck;
#if defined(CLK_TCK)
      clk_tck = CLK_TCK;
#else
      clk_tck = sysconf(_SC_CLK_TCK);
#endif

      times (&t);

      fprintf (stderr,
	     "\nRequested Allocated Freed     Attaches  Detaches  Frees     User    System\n");
      fprintf (stderr,
	     "%-9ld %-9ld %-9ld %-9ld %-9ld %-9ld %-7.2f %-7.2f\n",
	     requested, used, freed, attaches, detaches, frees,
	     0.01 * (float) ((t.tms_utime * 100) / clk_tck),
	     0.01 * (float) ((t.tms_stime * 100) / clk_tck));
      if (saved_calls)
	fprintf (stderr, "Saved %ld calls thru inlining.\n", saved_calls);
      if (array_ref)
	fprintf (stderr, "Saved %ld array references, left %ld.\n", array_ref, array_done);
    };
#endif
  exit (e);
}

void error (char *message)
{ fputs (message,stderr);
}
